#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <termios.h>
#include <unistd.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/**** sample: interval timer :begin ****/
#include <time.h>
#include <sys/time.h>
/**** sample: interval timer :end ****/

#define MAIN_C
#include "exitfail.h"

/**** sample: WD-3 serial setting: begin ****/
#define SERIAL_BAUDRATE B9600
/**** sample: WD-3 serial setting: end ****/
#define BUF_SIZE 2048

static int serial_fd = -1;
static struct termios old_tio;

static int terminated = 0;

#define BASENAME(p) ((strrchr((p), '/') ? : ((p) - 1)) +1)
#define ARRAY_SIZE(a) (sizeof(a) /sizeof((a)[0]))

#define __unused __attribute__((unused))

/**** sample: WD-3 command: begin ****/
static unsigned char cmd_start[] = "P"; /* set WD-3: start request*/
static unsigned char cmd_pwd[] = "ARP."; /* set WD-3: password */
static unsigned char cmd_mode[] = "M"; /* set WD-3: mode request*/
/**** sample: WD-3 command: end ****/

/**** sample: WD-3 response: begin ****/
static unsigned char res_ok[] = "O"; /* WD-3 response: ok */
/*static unsigned char res_ng[] = "N"; *//* WD-3 response: ng */
/**** sample: WD-3 response: end ****/

static void terminate_sig_handler(__unused int signum)
{
	terminated = 1;
}

static void set_sig_handler(int sig_list[], ssize_t num, __sighandler_t handler)
{
	struct sigaction sa;
	int i;

	memset(&sa, 0, sizeof(sa));
	sa.sa_handler = handler;

	for(i=0; i<num; i++)
		if(sigaction(sig_list[i], &sa, NULL) < 0)
			exitfail_errno("sigaction");
}

static void restore_serial(void)
{
	int ret;

	ret = tcsetattr(serial_fd, TCSANOW, &old_tio);
	if(ret < 0) exitfail_errno("tcsetattr");
}

static void setup_serial(int fd)
{
	struct termios tio;
	int ret;

	ret = tcgetattr(fd, &old_tio);
	if(ret) exitfail_errno("tcsgetattr");

	if(atexit(restore_serial)) exitfail_errno("atexit");

	memset(&tio, 0, sizeof(tio));
	tio.c_iflag = IGNBRK | IGNPAR;
	tio.c_cflag = CS8 | CLOCAL | CREAD;
	tio.c_cc[VTIME] = 0;
	tio.c_cc[VMIN] = 1;
	ret = cfsetspeed(&tio, SERIAL_BAUDRATE);
	if(ret < 0) exitfail_errno("cfsetspeed");

	ret = tcflush(fd, TCIFLUSH);
	if(ret < 0) exitfail_errno("tcflush");

	ret = tcsetattr(fd, TCSANOW, &tio);
	if(ret) exitfail_errno("tcsetattr");
}

int main(int argc, char *argv[])
{
	int terminate_sig_list[] = {
		SIGHUP, SIGINT, SIGQUIT, SIGPIPE, SIGTERM
	};

/**** sample: variable: begin ****/
	ssize_t ret;
       unsigned char buf[BUF_SIZE];
	int i;
	int flag_ok;
/**** sample: variable: end ****/

/**** sample: interval timer: begin ****/
	struct itimerval itimer;
/**** sample: interval timer: end ****/

	exitfail_init();

	if(argc < 2) {
		printf("Usage: %s <device>\n", BASENAME(argv[0]));
		return EXIT_SUCCESS;
	}

	serial_fd = open(argv[1], O_RDWR | O_NOCTTY);
	if(serial_fd < 0) exitfail_errno("open");

	set_sig_handler(terminate_sig_list, ARRAY_SIZE(terminate_sig_list),
			terminate_sig_handler);

	setup_serial(serial_fd);

/**** sample: interval timer: begin ****/
	itimer.it_value.tv_sec = itimer.it_interval.tv_sec = 5;
	itimer.it_value.tv_usec = itimer.it_interval.tv_usec = 0;
/**** sample: interval timer: end ****/

/**** sample: WD-3 start request: begin ****/
	ret = write(serial_fd, cmd_start, sizeof(cmd_start) - 1);
	if (ret < 0) exitfail_errno("write");
	printf("\n****************************************************\n");
	printf("-------- Command: %s", cmd_start);

	flag_ok = 0;

	while(!terminated && flag_ok != 1) {
		if(setitimer(ITIMER_REAL, &itimer, NULL) < 0) {
			exitfail_errno("setitimer");
			return 0;
		}
	
		ret = read(serial_fd, buf, BUF_SIZE);
		if(ret < 0) exitfail_errno("read");

		for(i = 0; i < ret && flag_ok != 1; i++) {
			printf("%c", buf[i]);
			if(buf[i] == res_ok[0]) flag_ok = 1;
		}
		if(flag_ok == 1) {
			printf("\n-------- Receive: %s\n", res_ok);
//			printf("%s\n", buf1);
		}
	}
/**** sample: WD-3 start request: end ****/
	sleep(3);
/**** sample: WD-3 password: begin ****/
	ret = write(serial_fd, cmd_pwd, sizeof(cmd_pwd) - 1);
	if (ret < 0) exitfail_errno("write");
	printf("\n****************************************************\n");
	printf("-------- Command: %s", cmd_pwd);

	flag_ok = 0;

	while(!terminated && flag_ok != 1) {
		if(setitimer(ITIMER_REAL, &itimer, NULL) < 0) {
			exitfail_errno("setitimer");
			return 0;
		}
	
		ret = read(serial_fd, buf, BUF_SIZE);
		if(ret < 0) exitfail_errno("read");

		for(i = 0; i < ret && flag_ok != 1; i++) {
			printf("%c", buf[i]);
			if(buf[i] == res_ok[0]) flag_ok = 1;
		}
		if(flag_ok == 1) {
			printf("\n-------- Receive: %s\n", res_ok);
//			printf("%s\n", buf1);
		}
	}
/**** sample: WD-3 password: end ****/
	sleep(3);
/**** sample: WD-3 mode request: begin ****/
	ret = write(serial_fd, cmd_mode, sizeof(cmd_mode) - 1);
	if (ret < 0) exitfail_errno("write");
	printf("\n****************************************************\n");
	printf("-------- Command: %s", cmd_mode);

	flag_ok = 0;

	while(!terminated && flag_ok != 1) {
		if(setitimer(ITIMER_REAL, &itimer, NULL) < 0) {
			exitfail_errno("setitimer");
			return 0;
		}
	
		ret = read(serial_fd, buf, BUF_SIZE);
		if(ret < 0) exitfail_errno("read");

		for(i = 0; i < ret && flag_ok != 1; i++) {
			printf("%c", buf[i]);
			if(buf[i] == res_ok[0]) flag_ok = 1;
		}
		if(flag_ok == 1) {
			printf("\n-------- Receive: %s\n", res_ok);
//			printf("%s\n", buf1);
		}
	}
/**** sample: WD-3 mode request: end ****/
	sleep(3);
	printf("\n######## Finish setting WD-3 #########\n\n");

	return EXIT_SUCCESS;
}

